/*
 * You aren't expected to be able to write code like this, but consider reading
 * through this file to get an idea of how event-driven execution works.
 */

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "list.h"

#define BUF_SIZE 1024
#define QUEUE_LENGTH 50

#define STATE_READ 0
#define STATE_WRITE 1

struct task {
  int fd;
  int state;
  char buf[BUF_SIZE];
  ssize_t buf_start;
  ssize_t buf_size;
  struct list_elem elem;
};
static struct list task_list;

struct addrinfo *setup_address(char *port) {
    struct addrinfo *server;
    struct addrinfo hints;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE;

    int rv = getaddrinfo(NULL, port, &hints, &server);
    if (rv != 0) {
        printf("getaddrinfo failed: %s\n", gai_strerror(rv));
        return NULL;
    }
    return server;
}

bool serve_client(struct task *task) {
    if (task->state == STATE_READ) {
        ssize_t n = read(task->fd, task->buf, BUF_SIZE - 1);
        if (n == 0 || n == -1) {
            if (n == -1) {
                if (errno == EAGAIN || errno == EWOULDBLOCK) {
                    /*
                     * If the buffer is empty, meaning we have to block to get
                     * data, we'll get this error. read() can't just return 0
                     * because otherwise it would look like EOF.
                     */
                    return true;
                }
                perror("read");
            }
            close(task->fd);
            return false;
        }
        task->buf_size = n;
        task->buf[task->buf_size] = '\0';
        task->buf_start = 0;
        task->state = STATE_WRITE; // echo what we got right away
        printf("Client Sent: %s\n", task->buf);
    } else if (task->state == STATE_WRITE) {
        ssize_t n = write(task->fd, &task->buf[task->buf_start], task->buf_size - task->buf_start);
        if (n == 0 || n == -1) {
            if (n == -1) {
                if (errno == EAGAIN || errno == EWOULDBLOCK) {
                    /*
                     * If the buffer is full, meaning we have to block to add
                     * data to it, we'll get this error.
                     */
                    return true;
                }
                perror("write");
            }
            close(task->fd);
            return false;
        }
        task->buf_start += n;
        if (task->buf_start == task->buf_size) {
            task->state = STATE_READ; // if we wrote everything, wait for next string from user
        }
    } else {
        abort(); // should not happen
    }
    return true;
}

int main(int argc, char **argv) {
    if (argc < 2) {
        printf("Usage: %s <port>\n", argv[0]);
        return 1;
    }
    list_init(&task_list);

    struct addrinfo *server = setup_address(argv[1]);
    if (server == NULL) {
        return 1;
    }
    int server_socket = socket(server->ai_family, server->ai_socktype, server->ai_protocol);
    if (server_socket == -1) {
        perror("socket");
        return 1;
    }
    /*
     * Mark the server socket as non-blocking. We don't want to "block"
     * waiting for a new connection
     */
    if (fcntl(server_socket, F_SETFL, O_NONBLOCK) == -1) {
        perror("fcntl");
        return 1;
    }
    if (bind(server_socket, server->ai_addr, server->ai_addrlen) == -1) {
        perror("bind");
        return 1;
    }

    freeaddrinfo(server);

    if (listen(server_socket, QUEUE_LENGTH) == -1) {
        perror("listen");
        return 1;
    }

    while (1) {
        /* Issue a select syscall to wait until a file descriptor is ready. */
        fd_set waiting_for_read;
        fd_set waiting_for_write;
        FD_ZERO(&waiting_for_read);
        FD_ZERO(&waiting_for_write);
        FD_SET(server_socket, &waiting_for_read);
        int max_fd = server_socket;
        for (struct list_elem *i = list_begin(&task_list); i != list_end(&task_list); i = list_next(i)) {
            struct task *curr = list_entry(i, struct task, elem);
            if (curr->state == STATE_READ) {
                FD_SET(curr->fd, &waiting_for_read);
            } else if (curr->state == STATE_WRITE) {
                FD_SET(curr->fd, &waiting_for_write);
            } else {
                abort(); // should not happen
            }
            max_fd = max_fd < curr->fd ? curr->fd : max_fd;
        }
        if (select(max_fd + 1, &waiting_for_read, &waiting_for_write, NULL, NULL) == -1) {
            perror("select");
            return 1;
        }

        /* We woke up. Check which FDs are ready and service them. */
        struct list_elem *i = list_begin(&task_list);
        while (i != list_end(&task_list)) {
            struct task *curr = list_entry(i, struct task, elem);
            i = list_next(i); // move to next elem now, in case we remove the current one below

            if (FD_ISSET(curr->fd, &waiting_for_read) || FD_ISSET(curr->fd, &waiting_for_write)) {
                bool still_alive = serve_client(curr);
                if (!still_alive) {
                    list_remove(&curr->elem);
                    free(curr);
                }
            }
        }

        if (FD_ISSET(server_socket, &waiting_for_read)) {
            int fd = accept(server_socket, NULL, NULL);
            if (fd == -1) {
                if (errno == EAGAIN || errno == EWOULDBLOCK) {
                    continue;
                }
                perror("accept");
                return 1;
            }
            if (fcntl(fd, F_SETFL, O_NONBLOCK) == -1) {
                perror("fcntl");
                return 1;
            }
            struct task *new_task = malloc(sizeof(struct task));
            if (new_task == NULL) {
                close(fd);
                continue;
            }
            new_task->fd = fd;
            new_task->state = STATE_READ;
            list_push_front(&task_list, &new_task->elem);
        }
    }

    return 0;
}
